from .base import Descriptor

__all__ = ['AccountFinancingMode', 'AccountID', 'PositionAggregationMode']


class AccountFinancingMode(Descriptor):
    """The financing mode of an Account
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'NO_FINANCING': 'No financing is paid/charged for open Trades in the Account',
        'SECOND_BY_SECOND': 'Second-by-second financing is paid/charged for open Trades in the Account, '
                            'both daily and when the the Trade is closed',
        'DAILY': 'A full day’s worth of financing is paid/charged for open Trades in the Account '
                 'daily at 5pm New York time'
    }


class AccountID(Descriptor):
    """The string representation of an Account Identifier.
    """

    # Type checking
    typ = str

    # Correct syntax of value
    format_syntax = '“-“-delimited string with format “{siteID}-{divisionID}-{userID}-{accountNumber}”'
    # Example of correct format
    example = '001-011-5838423-001'


class PositionAggregationMode(Descriptor):
    """The way that position values for an Account are calculated and aggregated.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'ABSOLUTE_SUM': 'The Position value or margin for each side (long and short) of '
                        'the Position are computed independently and added together.',
        'MAXIMAL_SIDE': 'The Position value or margin for each side (long and short) '
                        'of the Position are computed independently. The Position value or '
                        'margin chosen is the maximal absolute value of the two.',
        'NET_SUM': 'The units for each side (long and short) of the Position are netted '
                   'together and the resulting value (long or short) is used to compute '
                   'the Position value or margin.'
    }



from .base import Interface
from .decorators import endpoint
from ..definitions.types import CandlestickGranularity
from ..definitions.descriptors.primitives import InstrumentName
from ..definitions.types import Price
from ..definitions.types import WeeklyAlignment
from ..endpoints.annotations import AlignmentTimezone
from ..endpoints.annotations import Count
from ..endpoints.annotations import DailyAlignment
from ..endpoints.annotations import FromTime
from ..endpoints.annotations import Smooth
from ..endpoints.annotations import ToTime
from ..endpoints.annotations import includeFirst
from ..endpoints.instrument import *

__all__ = ['InstrumentInterface']


class InstrumentInterface(Interface):
    @endpoint(GETInstrumentsCandles)
    def get_candles(self,
                    instrument: InstrumentName,
                    price: Price,
                    granularity: CandlestickGranularity,
                    count: Count,
                    from_time: FromTime,
                    to_time: ToTime,
                    smooth: Smooth,
                    include_first: includeFirst,
                    daily_alignment: DailyAlignment,
                    alignment_timezone: AlignmentTimezone,
                    weekly_alignment: WeeklyAlignment,
                    ):
        """
        Fetch candlestick data for an instrument.

        Args:
            instrument:
                Name of the Instrument
            price:
                The Price component(s) to get candlestick data for. Can contain
                any combination of the characters "M" (midpoint candles) "B"
                (bid candles) and "A" (ask candles).
            granularity:
                The granularity of the candlesticks to fetch
            count:
                The number of candlesticks to return in the reponse. Count
                should not be specified if both the start and end parameters
                are provided, as the time range combined with the graularity
                will determine the number of candlesticks to return.
            from_time:
                The start of the time range to fetch candlesticks for.
            to_time:
                The end of the time range to fetch candlesticks for.
            smooth:
                A flag that controls whether the candlestick is "smoothed" or
                not.  A smoothed candlestick uses the previous candle's close
                price as its open price, while an unsmoothed candlestick uses
                the first price from its time range as its open price.
            include_first:
                A flag that controls whether the candlestick that is covered by
                the from time should be included in the results. This flag
                enables clients to use the timestamp of the last completed
                candlestick received to poll for future candlesticks but avoid
                receiving the previous candlestick repeatedly.
            daily_alignment:
                The hour of the day (in the specified timezone) to use for
                granularities that have daily alignments.
            alignment_timezone:
                The timezone to use for the dailyAlignment parameter.
                Candlesticks with daily alignment will be aligned to the
                dailyAlignment hour within the alignmentTimezone.
            weekly_alignment:
                The day of the week used for granularities that have weekly
                alignment.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass



from .base import Interface
from .decorators import endpoint
from ..definitions.types import DecimalNumber
from ..endpoints.account import *
from ..endpoints.annotations import *

__all__ = ['AccountInterface']


class AccountInterface(Interface):
    @endpoint(GETAccounts, serial=True)
    def list_accounts(self):
        """
        Get a list of all Accounts authorized for the provided token.

        Args:

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETAccountID, serial=True)
    def get_account_details(self):
        """
        Get the full details for a single Account that a client has access to.
        Full pending Order, open Trade and open Position representations are
        provided.

        Args:


        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETAccountIDSummary, serial=True)
    def account_summary(self):
        """
        Get a summary for a single Account that a client has access to.

        Args:


        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETAccountIDInstruments)
    def account_instruments(self, instruments: Instruments):
        """
        Get the list of tradeable instruments for the given Account. The list
        of tradeable instruments is dependent on the regulatory division that
        the Account is located in, thus should be the same for all Accounts
        owned by a single user.

        Args:

            instruments:
                List of instruments to query specifically.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(PATCHAccountIDConfiguration)
    def configure_account(self, alias: Alias, margin_rate: DecimalNumber):
        """
        Set the client-configurable portions of an Account.

        Args:

            alias:
                Client-defined alias (name) for the Account
            margin_rate:
                The string representation of a decimal number.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETAccountIDChanges)
    def account_changes(self, since_transaction_id: LastTransactionID):
        """
        Endpoint used to poll an Account for its current state and changes
        since a specified TransactionID.

        Note:
            since_transaction_id is automatically handled by OandaClient

        Args:
            since_transaction_id:
                ID of the Transaction to get Account changes since.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass



import os
from functools import partial

from yarl import URL

from .definitions.types import AcceptDatetimeFormat
from .endpoints.annotations import Authorization
from .helpers import request_limiter, initializer
from .interface import *

version = '1.1.5a4'


class OandaClient(AccountInterface, InstrumentInterface, OrderInterface, PositionInterface,
                  PricingInterface,
                  TradeInterface,
                  TransactionInterface, UserInterface):
    """
    Create an API context for v20 access

    Args:
        token: -- User generated token from the online account configuration page
        rest_host: -- The hostname of the v20 REST server
        rest_port: -- The port of the v20 REST server
        stream_host: -- The hostname of the v20 REST server
        stream_port: -- The port of the v20 REST server
        application: Optional name of the application using the v20 bindings
        datetime_format: -- The format to request when dealing with times
        poll_timeout: -- The timeout to use when making a polling request with
            the v20 REST server
        max_requests_per_second: -- Maximum HTTP requests sent per second
        max_simultaneous_connections: -- Maximum concurrent HTTP requests

    """

    default_parameters = {}

    initialized = 0

    account = None

    session = None

    loop = None

    def __init__(self, token=None, rest_host='api-fxpractice.oanda.com', rest_port=443,
                 stream_host='stream-fxpractice.oanda.com', stream_port=None, application='async_v20',
                 datetime_format='UNIX', poll_timeout=2, max_requests_per_second=99, max_simultaneous_connections=10):

        # TODO: add poll timeout
        self.version = version

        if token is None:
            token = os.environ['OANDA_TOKEN']

        self.application = application

        # V20 REST API URL
        rest_host = partial(URL.build, host=rest_host, port=rest_port, scheme='https')

        # v20 STREAM API URL
        stream_host = partial(URL.build, host=stream_host, port=stream_port, scheme='https')

        self.hosts = {'REST': rest_host, 'STREAM': stream_host}

        # The timeout to use when making a polling request with the
        # v20 REST server
        self.poll_timeout = poll_timeout

        # TODO limit this to > 0
        # Limit new requests to a certain rate
        self.max_requests_per_second = max_requests_per_second

        # TODO limit this to > 0
        # Limit concurrent connections
        self.max_simultaneous_connections = max_simultaneous_connections

        # This is the default parameter dictionary. OandaClient Methods that require certain parameters
        # that are  not explicitly passed will try to find it in this dict
        self.default_parameters.update(
            {Authorization: 'Bearer {}'.format(token),
             AcceptDatetimeFormat: datetime_format}
        )

        self.request = request_limiter(self)

        self.initialize_client = initializer(self)

    async def initialize(self):
        await self.initialize_client.asend(None)

    async def __aenter__(self):
        await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __enter__(self):
        # TODO Make this print in red
        print('Warning: `with` used rather than `async with`')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.session.close()




from .base import Descriptor

__all__ = ['AcceptDatetimeFormat', 'AccountUnits', 'Currency', 'DateTime', 'DecimalNumber',
           'Unit', 'Direction',
           'InstrumentName', 'InstrumentType']


class AcceptDatetimeFormat(Descriptor):
    """DateTime header
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'UNIX': 'If “UNIX” is specified DateTime fields will be specified or '
                'returned in the “12345678.000000123” format.',
        'RFC3339': 'If “RFC3339” is specified DateTime will be specified or '
                   'returned in “YYYY-MM-DDTHH:MM:SS.nnnnnnnnnZ” format.'
    }


class AccountUnits(Descriptor):
    """The string representation of a quantity of an Account’s home currency.
    """

    # Type checking
    # TODO keep an eye on this. OANDA specifies this as a str.
    # Though it makes more sense for it to be a float
    # floats automatically get converted to to strings anyway
    # when serialized into JSON
    typ = float

    # Correct syntax of value
    format_syntax = 'A decimal number encoded as a string. The amount of precision ' \
                    'provided depends on the Account’s home currency.'


class Currency(Descriptor):
    """Currency name identifier. Used by clients to refer to currencies.
    """

    # Type checking
    typ = str

    # Correct syntax of value
    format_syntax = 'A string containing an ISO 4217 currency'


class DateTime(Descriptor):
    """A date and time value using either RFC3339 or UNIX time representation.
    """

    # Type checking
    typ = str

    # Correct syntax of value
    format_syntax = 'The RFC 3339 representation is a string conforming to'


class DecimalNumber(Descriptor):
    """The string representation of a decimal number.
    """

    # Type checking
    typ = float

    # Correct syntax of value
    format_syntax = 'A decimal number encoded as a string. The amount of precision ' \
                    'provided depends on what the number represents.'

    def __set__(self, instance, value):
        value = super().type_check(value)
        super().__set__(instance, round(value, 5))

class Unit(Descriptor):
    """A unit is a standard allotment of a currency
    """

    # Type checking
    typ = float

    # Correct syntax of value
    format_syntax = 'A decimal number encoded as a string. The amount of precision ' \
                    'provided depends on what the number represents.'

    def __set__(self, instance, value):
        value = super().type_check(value)
        super().__set__(instance, round(value, 0))


class Direction(Descriptor):
    """In the context of an Order or a
    Trade, defines whether the units are positive or negative.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'LONG': 'A long Order is used to to buy units of an Instrument. '
                'A Trade is long when it has bought units of an Instrument.',
        'SHORT': 'A short Order is used to to sell units of an Instrument.'
                 'A Trade is short when it has sold units of an Instrument.'
    }


class InstrumentName(Descriptor):
    """Instrument name identifier. Used by clients to refer to an Instrument.
    """

    # Type checking
    typ = str

    # Correct syntax of value
    format_syntax = 'A string containing the base currency and quote currency delimited by a “_”.'


class InstrumentType(Descriptor):
    """The type of an Instrument.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'CURRENCY': 'Currency',
        'CFD': 'Contract For Difference',
        'METAL': 'Metal'
    }



import ujson as json

import pandas as pd

from .metaclass import *


class Model(metaclass=ORM):
    _schema = {}
    _fields = []
    _delimiter = '_'
    _derived = None

    # More info about this code be found in PEP 487 https://www.python.org/dev/peps/pep-0487/
    def __init_subclass__(cls, **kwargs):
        dispatch_key = cls._schema.get('type', None)
        if dispatch_key:
            cls._dispatch.update({dispatch_key.default: cls})
        else:
            cls._derived = cls

    def __new__(cls, *args, **kwargs):
        if cls._dispatch:
            typ = kwargs.get('type', None)
            if typ:
                return cls._dispatch[typ](*args, **kwargs)
        else:
            return super().__new__(cls)

    def __repr__(self):
        return self.__class__.__name__

    # Recursion might be an issue
    def json_dict(self, float_to_string=True):
        def get_object_fields(attr):
            attr = getattr(self, attr)
            if not isinstance(attr, (int, float, str)):
                try:
                    attr = attr.json_dict(float_to_string)
                except AttributeError:
                    attr = [obj.json_dict(float_to_string) for obj in attr]
            elif float_to_string and isinstance(attr, float):
                attr = str(attr)
            return attr

        return {self.__class__.json_attributes[attr]: get_object_fields(attr)
                for attr in self._fields}


    def json_data(self):
        return json.dumps(self.json_dict(float_to_string=True))

    def data(self, float_to_string=False):
        return flatten_dict(self.json_dict(float_to_string), self._delimiter)

    def series(self):

        def str_to_int(value):
            try:
                value = int(value)
            except ValueError:
                pass
            return value

        data = {key: str_to_int(value) if isinstance(value, str) else value
                for key, value in self.data(float_to_string=False).items()}
        return pd.Series(dict(self.template, **data))



from .base import Descriptor

__all__ = ['TradeID', 'TradePL', 'TradeSpecifier', 'TradeState', 'TradeStateFilter']


class TradeID(Descriptor):
    """The Trade’s identifier, unique within the Trade’s Account.
    """

    # Type checking
    typ = int

    # Correct syntax of value
    format_syntax = 'The string representation of the OANDA-assigned TradeID. ' \
                    'OANDA-assigned TradeIDs are positive integers, and are ' \
                    'derived from the TransactionID of the Transaction that opened the Trade.'
    # # Example of correct format
    # example = '1523'


class TradePL(Descriptor):
    """The classification of TradePLs.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'POSITIVE': 'An open Trade currently has a positive (profitable) unrealized P/L, '
                    'or a closed Trade realized a positive amount of P/L.',
        'NEGATIVE': 'An open Trade currently has a negative (losing) unrealized P/L, '
                    'or a closed Trade realized a negative amount of P/L.',
        'ZERO': 'An open Trade currently has unrealized P/L of zero (neither profitable nor losing),'
                ' or a closed Trade realized a P/L amount of zero.'
    }


class TradeSpecifier(Descriptor):
    """The identification of a Trade as referred to by clients
    """

    # Type checking
    typ = str

    # Correct syntax of value
    format_syntax = 'Either the Trade’s OANDA-assigned TradeID or the Trade’s client-provided ' \
                    'ClientID prefixed by the “@” symbol'
    # Example of correct format
    example = '@my_trade_id'


class TradeState(Descriptor):
    """The current state of the Trade.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'OPEN': 'The Trade is currently open',
        'CLOSED': 'The Trade has been fully closed',
        'CLOSE_WHEN_TRADEABLE': 'The Trade will be closed as soon as the trade’s instrument becomes tradeable'
    }


class TradeStateFilter(Descriptor):
    """The state to filter the Trades by
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'OPEN': 'The Trades that are currently open',
        'CLOSED': 'The Trades that have been fully closed',
        'CLOSE_WHEN_TRADEABLE': 'The Trades  that will be closed as soon as the trades’ instrument becomes tradeable',
        'ALL': 'The Trades that are in any of the possible states listed above.'
    }




from inspect import Signature, Parameter, _empty
from itertools import chain, starmap

from inflection import underscore

from .descriptors.base import Descriptor


def flatten_dict(dictionary, delimiter='_'):
    """Flatten a nested dictionary structure"""

    def unpack(parent_key, parent_value):
        """Unpack one level of nesting in a dictionary"""
        try:
            items = parent_value.items()
        except AttributeError:
            # parent_value was not a dict, no need to flatten
            yield (parent_key, parent_value)
        else:
            for key, value in items:
                yield (parent_key + delimiter + key, value)

    while True:
        # Keep unpacking the dictionary until all value's are not dictionary's
        dictionary = dict(chain.from_iterable(starmap(unpack, dictionary.items())))
        if not any(isinstance(value, dict) for value in dictionary.values()):
            break

    return dictionary


def create_signature(cls):
    schema = cls._schema

    def create_parameter(key, schema_value):
        name = cls.instance_attributes[key]
        annotation = schema_value.typ
        default = schema_value.default
        if default == _empty and schema_value.required is False:
            default = None
        return Parameter(name=name, annotation=annotation, default=default, kind=Parameter.POSITIONAL_OR_KEYWORD)

    def sort_key(param):
        default = False
        if param.default != _empty:
            default = True
        return default

    def parameters(schema):
        yield Parameter(name='self', kind=Parameter.POSITIONAL_ONLY)
        for key, value in schema.items():
            yield create_parameter(key, value)

    return Signature(sorted(parameters(schema), key=sort_key))


def create_instance_attributes(cls):
    instance_attributes = {key: underscore(key) for key in cls._schema}
    instance_attributes.update({value: value for value in instance_attributes.values()})
    return instance_attributes


def create_json_attributes(cls):
    return {underscore(key): key for key in cls._schema}


def assign_descriptors(cls):
    for attr, schema_value in cls._schema.items():
        typ = schema_value.typ
        if issubclass(typ, Descriptor):
            attr = cls.instance_attributes[attr]
            setattr(cls, attr, typ(name='_' + attr))
    return cls


def create_attribute(typ, data):
    if isinstance(data, dict):
        result = typ(**data)
    else:
        result = typ(data)
    return result



import asyncio

from async_v20 import OandaClient


async def candles():
    async with OandaClient() as client:
        return await client.get_candles('AUD_USD', granularity='M')


loop = asyncio.get_event_loop()
response = loop.run_until_complete(candles())

# HTTP response state
print(response['instrument'])

# JSON data in python dictionary format
print(response['candles'][0].json_dict())

# pandas Series
print(response['candles'][0].series())